/*
 * FILE: librenewssp.c
 *
 * Comment: RenewSSP implementation as a shared library.
 *          Use it by preloading this library when launching an executable.
 *          
 *          $ export LD_PRELOAD=./librenewssp.so
 *          $ ./test_renewssp
 *          Canary before fork():           0x31f1b82932a68800
 *          Reference-canary of the parent: 0x31f1b82932a68800
 *          Reference-canary of the child:  0x47f0294b388f0200
 *
 * Author: Hector Marco <hecmargi@upv.es>
 *         Ismael Ripoll <iripoll@disca.upv.es>
 *
 * Date:   February 2014
 *
 * Web:    http://renewssp.com    and    http://cybersecurity.upv.es
 * 
 * (C) Copyright, 2014  Hector Marco and Ismael Ripoll
 * 
 * This Software is provided to you solely for testing and
 * evaluation. You may not use it for any other purpose.
 *
 * The software is provided "as is", without warranty of any kind,
 * express or implied, including but not limited to the warranties of
 * merchantability, fitness for a particular purpose and
 * noninfringement. In no event shall the authors or copyright holders
 * be liable for any claim, damages or other liability, whether in an
 * action of contract, tort or otherwise, arising from, out of or in
 * connection with the software or the use or other dealings in the
 * software.
 *
 */ 



#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdint.h>

#define __USE_GNU
#include <dlfcn.h>


#ifdef __i386__
# define THREAD_SET_STACK_GUARD(x)			\
	asm volatile ("mov %0, %%gs:0x14" ::"r" (x) : "memory");
#elif defined __x86_64__
# define THREAD_SET_STACK_GUARD(x)			\
	asm volatile ("mov %0, %%fs:0x28" ::"r" (x) : "memory");
#endif

static pid_t (*real_fork)(void)=NULL;

static void renew_randomly_stack_chk_guard(void) {
	union   {
		uintptr_t num;
		unsigned char bytes[sizeof (uintptr_t)];
	} ret;
	ssize_t reslen;
	const ssize_t filllen = sizeof (ret.bytes) - 1;
	
	ret.num = 0;
	int fd = open ("/dev/urandom", O_RDONLY);
	if ( fd >= 0 ) {
		reslen = read (fd, ret.bytes + 1, filllen);
		close (fd);
		if (reslen != filllen) {
			fprintf(stderr, "Warning: renewssp was not able to renew the canary.\n");
		}
	}
	
	THREAD_SET_STACK_GUARD(ret.num);
}


static void __raf_fork_init(void) {
	
	real_fork = dlsym(RTLD_NEXT, "fork");
	if (NULL == real_fork) {
		fprintf(stderr, "Warning: renewssp can not overload `dlsym`: %s\n", dlerror());
	}
}

pid_t fork(void) {
	pid_t pid;
	
	if (real_fork==NULL) {
		__raf_fork_init();
	}
	
	pid = real_fork();
	
	if (pid == 0) {
		/* Only renew the reference canary in the child process. */
		renew_randomly_stack_chk_guard();
	}
       
	return pid;
}
